import javax.microedition.lcdui.*;
import java.util.*;

public class BrowserCanvas extends Canvas implements CommandListener {
    static final int IMAGE_SIZE = 256;
    static final int IMAGE_CACHE_SIZE = 4;
    static final int IMAGE_WIDTH = 4494;
    static final int IMAGE_HEIGHT = 3179;

    static final int INIT_X = IMAGE_WIDTH/2;
    static final int INIT_Y = IMAGE_HEIGHT/2;

    static final int BLINK_PERIOD = 1000;
    static final int REPEAT_DELAY = 1000;
    static final int REPEAT_PERIOD = 100;

    static final int BACKGROUND_COLOR = 0xffffff;
    static final int SELECTION_COLOR = 0x800080;
    static final int FOCUS_COLOR = 0x000000;
    static final int EPSILON = 10;
    static final int SCALE = 10;

    int x, y;
    int focusX, focusY;
    int focusWidth, focusHeight;
    int focusX0, focusY0;
    int screenWidth, screenHeight;
    int selectionColor = SELECTION_COLOR;
    MapBrowser browser;
    int scale;

    Image   allImage;
    Image[] imageCache = new Image[IMAGE_CACHE_SIZE];
    int[]   imagePos = new int[IMAGE_CACHE_SIZE];
    int     imageCachePtr;
    int     stepX, stepY;
    int     focusStepX, focusStepY;
    Timer   repeatTimer = new Timer();
    Timer   blinkTimer = new Timer();
    RepeatTask repeatTask;
    BlinkTask  blinkTask;

    Location[] selection;
    Index      selectionIndex;
    int        currSelection;

    boolean    nokia6600;

    void error(String text, final Displayable parent) {
        Alert a = new Alert(Locale.current.getResource("Error"));
        a.setType(AlertType.ERROR);
        a.setString(text);
        a.setTimeout(Alert.FOREVER);
        if (nokia6600) { 
            Display.getDisplay(browser).setCurrent(parent);
            Display.getDisplay(browser).setCurrent(a, parent);
            a.setCommandListener(new CommandListener() { 
                    public void commandAction(Command c, Displayable d) { 
                        Display.getDisplay(browser).setCurrent(parent);
                    }
                }
            );
        } else { 
            Display.getDisplay(browser).setCurrent(a, parent);        
        }         
    }

    Image getImage(int x, int y) { 
        int pos = (x << 16) + y + 1;
        for (int i = 0; i < IMAGE_CACHE_SIZE; i++) {
            if (imagePos[i] == pos) { 
                return imageCache[i];
            }
        }
        String imageName = "/city" + y + '_' + x + ".png";
        //String imageName = "/img" + (y+1) + '_' + (x+1) + ".png";
        try { 
            Image image = Image.createImage(imageName);
            imageCache[imageCachePtr] = image;
            imagePos[imageCachePtr] = pos;
            imageCachePtr = (imageCachePtr + 1) % IMAGE_CACHE_SIZE;
            return image;
        } catch (Exception e) { 
            System.err.println("Failed to load image " + imageName);
            return null;
        }
    }

    void setPositions(Index index, Location[] locations) { 
        selection = locations;
        selectionIndex = index;
        currSelection = -1;
        if (locations.length == 1) {
            Location l = locations[0];
            if ((browser.settings.flags & Settings.AUTO_ZOOM) != 0) { 
                if (l.getWidth() > screenWidth || l.getHeight() > screenHeight) {             
                    scale = SCALE;
                    addCommand(MapBrowser.ZOOM_IN_CMD);
                    removeCommand(MapBrowser.ZOOM_OUT_CMD);
                    x = l.getCenterX() - screenWidth*SCALE/2;
                    y = l.getCenterY() - screenHeight*SCALE/2;
                    focusX = x + focusX0;
                    focusY = y + focusY0;
                } else { 
                    scale = 1;
                    addCommand(MapBrowser.ZOOM_OUT_CMD);
                    removeCommand(MapBrowser.ZOOM_IN_CMD);
                    x = l.getCenterX() - screenWidth/2;
                    y = l.getCenterY() - screenHeight/2;
                }            
            } else { 
                x = l.getCenterX() - screenWidth*scale/2;
                y = l.getCenterY() - screenHeight*scale/2;
                focusX = x + focusX0;
                focusY = y + focusY0;
            }
        }
        if (locations.length > 0) { 
            if ((browser.settings.flags & Settings.BLINKING) != 0) { 
                startBlinking();
            }
            if (index.type == Index.POINT) { 
                addCommand(MapBrowser.NEXT_CMD);
            }
            addCommand(MapBrowser.CLEAN_CMD);
        }
        check();
    }

    void startBlinking() { 
        blinkTask = new BlinkTask();
        blinkTimer.schedule(blinkTask, BLINK_PERIOD, BLINK_PERIOD);
    }

    void stopBlinking() { 
        if (blinkTask != null) { 
            blinkTask.cancel();
            blinkTask = null;
        }
    }

    BrowserCanvas(MapBrowser browser) { 
        this.browser = browser;
        setCommandListener(this);
        addCommand(MapBrowser.INDEX_CMD);
        addCommand(MapBrowser.SETTINGS_CMD);
        addCommand(MapBrowser.HELP_CMD);
        addCommand(MapBrowser.ABOUT_CMD);
        addCommand(MapBrowser.QUIT_CMD);
        screenWidth = getWidth();
        screenHeight = getHeight();
        focusX0 = screenWidth*(SCALE-1)/2;
        focusY0 = screenHeight*(SCALE-1)/2;
        focusWidth = screenWidth/SCALE;
        focusHeight = screenHeight/SCALE;
        focusStepX = focusWidth/SCALE;
        focusStepY = focusHeight/SCALE;
        stepX = screenWidth/4;
        stepY = screenHeight/4;
        try { 
            allImage = Image.createImage("/all.png");
        } catch (Exception e) { 
            System.err.println("Failed to load image all.png");
        }     
        if ((browser.settings.flags & Settings.ZOOM_IN) != 0) { 
            scale = 1;
            addCommand(MapBrowser.ZOOM_OUT_CMD);
        } else { 
            scale = SCALE;
            addCommand(MapBrowser.ZOOM_IN_CMD);
        }
        x = browser.settings.x;
        y = browser.settings.y;
        focusX = x + focusX0;
        focusY = y + focusY0;
        
        nokia6600 = "Nokia6600".equals(System.getProperty("microedition.platform"));
        Display.getDisplay(browser).setCurrent(this);
    }
        
    protected void paint(Graphics g) {
        //System.out.println("Repaint xc=" + g.getClipX() + ", yc=" + g.getClipY() + ", width=" + g.getClipWidth() + ", height=" + g.getClipHeight() + ", x=" + x + ", y=" + y + ", scale=" + scale);
        g.setColor(BACKGROUND_COLOR);
        g.fillRect(g.getClipX(), g.getClipY(), g.getClipWidth(), g.getClipHeight());
        //System.out.println("Draw x=" + x + ", y=" + y + ", centerX = " + (x + screenWidth*scale/2) + 
        //                   ", centerY=" + (y + screenHeight*scale/2) + ", scale=" + scale);
        if (scale == 1) { 
            int x0 = x + g.getClipX();
            int y0 = y + g.getClipY();
            int x1 = x0 + g.getClipWidth() - 1;
            int y1 = y0 + g.getClipHeight() - 1;
            for (int i = x0/IMAGE_SIZE; i <= x1/IMAGE_SIZE; i++) { 
                for (int j = y0/IMAGE_SIZE; j <= y1/IMAGE_SIZE; j++) { 
                    Image image = getImage(i, j);
                    if (image != null) { 
                        g.drawImage(image, i*IMAGE_SIZE-x, j*IMAGE_SIZE-y, Graphics.TOP|Graphics.LEFT);
                    }
                }
            }
        } else { 
            g.drawImage(allImage, -x/SCALE, -y/SCALE, Graphics.TOP|Graphics.LEFT);
            g.setColor(FOCUS_COLOR);
            g.setStrokeStyle(Graphics.DOTTED);
            g.drawRect((focusX-x)/SCALE, (focusY-y)/SCALE, focusWidth, focusHeight);
        }
        if (selection != null) { 
            g.setStrokeStyle(Graphics.DOTTED);
            g.setColor(selectionColor);
            for (int i = 0; i < selection.length; i++) { 
                Location loc = selection[i];
                switch (selectionIndex.type) { 
                case Index.POINT:
                    g.drawLine((loc.points[0]-x)/scale-EPSILON, (loc.points[1]-y)/scale, 
                               (loc.points[0]-x)/scale+EPSILON, (loc.points[1]-y)/scale);
                    g.drawLine((loc.points[0]-x)/scale, (loc.points[1]-y)/scale-EPSILON, 
                               (loc.points[0]-x)/scale, (loc.points[1]-y)/scale+EPSILON);
                    if (i == currSelection) { 
                        g.drawArc((loc.points[0]-x)/scale-EPSILON, (loc.points[1]-y)/scale-EPSILON,
                                  EPSILON*2, EPSILON*2, 0, 360);
                    }
                    break;
                case Index.RECTANGLE:
                    g.drawRect((loc.points[0]-x)/scale, (loc.points[1]-y)/scale, 
                               (loc.points[2]-loc.points[0])/scale+1, 
                               (loc.points[3]-loc.points[1])/scale+1);
                    break;
                case Index.LINE:
                    for (int j = 2; j < loc.points.length; j += 2) { 
                        g.drawLine((loc.points[j-2]-x)/scale, (loc.points[j-1]-y)/scale, 
                                   (loc.points[j]-x)/scale, (loc.points[j+1]-y)/scale);
                    }
                    break;
                }
            }
        }
    }    

    void zoomIn() { 
        removeCommand(MapBrowser.ZOOM_IN_CMD);
        addCommand(MapBrowser.ZOOM_OUT_CMD);
        x = focusX;
        y = focusY;
        scale = 1;
    }        

    void zoomOut() { 
        removeCommand(MapBrowser.ZOOM_OUT_CMD);
        addCommand(MapBrowser.ZOOM_IN_CMD);
        focusX = x;
        focusY = y;
        x -= focusX0;
        y -= focusY0;
        scale = SCALE;
    }        

    void zoom() { 
        if (scale == 1) { 
            zoomOut();
        } else { 
            zoomIn();
        }
    }        

    public void commandAction(Command c, Displayable d)  
    {
        if (c == MapBrowser.QUIT_CMD) {
            browser.quit();
        } else if (c == MapBrowser.ABOUT_CMD) {
            new HelpForm(browser, Locale.current.getResource("About"), Locale.current.getResource("AboutText"), this);
        } else if (c == MapBrowser.HELP_CMD) {
            new HelpForm(browser, Locale.current.getResource("Help"), Locale.current.getResource("HelpText"), this);
        } else if (c == MapBrowser.ZOOM_IN_CMD) {
            zoomIn();
            check();
        } else if (c == MapBrowser.ZOOM_OUT_CMD) {
            zoomOut();
            check();
        } else if (c == MapBrowser.INDEX_CMD) {
            new GroupList(this);
        } else if (c == MapBrowser.NEXT_CMD) {
            nextPoint();
            repaint();
        } else if (c == MapBrowser.INFO_CMD) {
            pointInfo();
        } else if (c == MapBrowser.CLEAN_CMD) {
            selection = null;
            stopBlinking();
            removeCommand(MapBrowser.CLEAN_CMD);
            removeCommand(MapBrowser.NEXT_CMD);
            removeCommand(MapBrowser.INFO_CMD);            
        } else if (c == MapBrowser.SETTINGS_CMD) {
            new SettingsForm(this);
        }
    }
 
    class RepeatTask extends TimerTask { 
        RepeatTask(int keyCode) { 
            this.keyCode = keyCode;
        }

        public void run() { 
            keyAction(keyCode);
        }

        int keyCode;
    }

    class BlinkTask extends TimerTask { 
        public void run() { 
            selectionColor ^= 0xffffff;
            repaint();
        }
    }

    public void keyPressed(int keyCode) {
        if (repeatTask != null) { 
            repeatTask.cancel();
        }
        repeatTask = new RepeatTask(keyCode);
        repeatTimer.schedule(repeatTask, REPEAT_DELAY, REPEAT_PERIOD);
    }

    public void keyReleased(int keyCode) {
        keyAction(keyCode);
        if (repeatTask != null) { 
            repeatTask.cancel();
	    repeatTask = null;
        }
    }
    
    private boolean pointInfo() { 
        if (selection != null && currSelection >= 0) { 
            Location loc = selection[currSelection];
            new InfoForm(this, selectionIndex, selection[currSelection], this);
            return true;
        }        
        return false;
    }

    private boolean nextPoint() { 
        if (selection != null && selectionIndex.type == Index.POINT) { 
            for (int i = 0; i < selection.length; i++) { 
                if (++currSelection == selection.length) { 
                    currSelection = 0;
                }
                Location loc = selection[currSelection];
                if (loc.points[0] >= x && loc.points[0] - x < screenWidth*scale
                    && loc.points[1] >= y && loc.points[1] - y < screenHeight*scale)
                {
                    addCommand(MapBrowser.INFO_CMD);
                    return true;
                }
            }
            currSelection = -1;
        }
        return false;
    }

    private void keyAction(int keyCode) {
        int dx = 0, dy = 0;

	switch (keyCode) {
        case KEY_NUM0:
            focusX = x + focusX0;
            focusY = y + focusX0;
            repaint();
            return;
        case KEY_NUM1:
            dx = -1;
            dy = -1;
            break;
        case KEY_NUM2:
            dy = -1;
            break;
        case KEY_NUM3:
            dx = 1;
            dy = -1;
            break;
        case KEY_NUM4:
            dx = -1;
            break;
        case KEY_NUM5:
            zoom();
            check();
            return;
        case KEY_NUM6:
            dx = 1;
            break;
        case KEY_NUM7:
            dx = -1;
            dy = 1;
            break;
        case KEY_NUM8:
            dy = 1;
            break;
        case KEY_NUM9:
            dx = 1;
            dy = 1;
            break;
        case KEY_STAR:
            if (!nextPoint()) { 
                error(Locale.current.getResource("No points selected"), this);
            }
            repaint();
            return;
        case KEY_POUND:
            if (!pointInfo()) { 
                error(Locale.current.getResource("No current point"), this);
            }
            return;
        default:
            if ((browser.settings.flags & Settings.FAST_SCROLL) != 0) { 
                dx = stepX*scale;
                dy = stepY*scale;
                switch (getGameAction(keyCode)) { 
                case LEFT:
                    x -= dx;
                    break;
                case RIGHT:
                    x += dx;
                    break;
                case UP:
                    y -= dy;
                    break;
                case DOWN:
                    y += dy;
                    break;
                case FIRE:
                    zoom();
                    break;
                default:
                    return;
                }
                check();
                return;
            } else { 
                switch (getGameAction(keyCode)) {
                case UP:
                    dy = -1;
                    break;
                case LEFT:
                    dx = -1;
                    break;
                case FIRE:
                    zoom();
                    check();
                    return;
                case RIGHT:
                    dx = 1;
                    break;
                case DOWN:
                    dy = 1;
                    break;
                }
            }
        }
        if (scale != 1) { 
            boolean repaintAll = false;
            focusX += dx*focusWidth;
            focusY += dy*focusHeight;
            if (focusX < x) { 
                repaintAll = true;
                x -= stepX*SCALE;
                focusX += stepX*SCALE;
                if (x < 0) { 
                    focusX += x;
                    x = 0;
                    if (focusX < 0) { 
                        focusX = 0;
                    }
                }
            }
            if (focusX - x > (screenWidth - focusWidth)*SCALE) { 
                repaintAll = true;
                x += stepX*SCALE;
                focusX -= stepX*SCALE;
                if (x + screenWidth*SCALE > IMAGE_WIDTH) { 
                    focusX += x + screenWidth*SCALE - IMAGE_WIDTH;
                    x = IMAGE_WIDTH - screenWidth*SCALE;
                    if (focusX + focusWidth*SCALE > IMAGE_WIDTH) { 
                        focusX = IMAGE_WIDTH - focusWidth*SCALE;
                    }
                }
            }
            if (focusY < y) { 
                repaintAll = true;
                y -= stepY*SCALE;
                focusY += stepY*SCALE;
                if (y < 0) { 
                    focusY += y;
                    y = 0;
                    if (focusY < 0) { 
                        focusY = 0;
                    }
                }
            }
            if ((focusY - y) > (screenHeight - focusHeight)*SCALE) { 
                repaintAll = true;
                y += stepY*SCALE;
                focusY -= stepY*SCALE;
                if (y + screenHeight*SCALE > IMAGE_HEIGHT) { 
                    focusY += y + screenHeight*SCALE - IMAGE_HEIGHT;
                    y = IMAGE_HEIGHT - screenHeight*SCALE;
                    if (focusY + focusHeight*SCALE > IMAGE_HEIGHT) { 
                        focusY = IMAGE_HEIGHT - focusHeight*SCALE;
                    }
                }
            }
            if (repaintAll) { 
                repaint();
            } else { 
                repaint(((dx <= 0 ? focusX : focusX - focusWidth) - x) / SCALE,
                        ((dy <= 0 ? focusY : focusY - focusHeight) - y) / SCALE,
                        focusWidth + focusWidth/SCALE + 2, 
                        focusHeight + focusHeight/SCALE + 2);
            }
        } else { 
            x += dx*stepX;
            y += dy*stepY;
            check();
        }
    }


    private void check() { 
        if (x < 0) { 
            x = 0;
        } else if (x + screenWidth*scale > IMAGE_WIDTH) { 
            x = IMAGE_WIDTH - screenWidth*scale;
        }
        if (y < 0) { 
            y = 0;
        } else if (y + screenHeight*scale > IMAGE_HEIGHT) { 
            y = IMAGE_HEIGHT - screenHeight*scale;
        }
        if (scale != 1 && 
            (focusX < x || (focusX - x) > (screenWidth - focusWidth)*SCALE
             || focusY < y || (focusY - y) > (screenHeight - focusHeight)*SCALE))
        {
            focusX = x + focusX0;
            focusY = y + focusY0;
        }
        repaint();
    }

}
